# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Finds components for a given manifest."""

import pathlib
import sys
from typing import Collection, Container
import xml.etree.ElementTree


def _element_is_compatible_with_device_core(
    element: xml.etree.ElementTree.Element, device_core: str | None
) -> bool:
    """Check whether element is compatible with the given core.

    Args:
        element: element to check.
        device_core: name of core to filter sources for.

    Returns:
        True if element can be used, False otherwise.
    """
    if device_core is None:
        return True

    value = element.attrib.get('device_cores', None)
    if value is None:
        return True

    device_cores = value.split()
    return device_core in device_cores


def get_component(
    root: xml.etree.ElementTree.Element,
    component_id: str,
    device_core: str | None = None,
) -> tuple[xml.etree.ElementTree.Element | None, pathlib.Path | None]:
    """Parse <component> manifest stanza.

    Schema:
        <component id="{component_id}" package_base_path="component"
                   device_cores="{device_core}...">
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.
        device_core: name of core to filter sources for.

    Returns:
        (element, base_path) for the component, or (None, None).
    """
    xpath = f'./components/component[@id="{component_id}"]'
    component = root.find(xpath)
    if component is None or not _element_is_compatible_with_device_core(
        component, device_core
    ):
        return (None, None)

    try:
        base_path = pathlib.Path(component.attrib['package_base_path'])
        return (component, base_path)
    except KeyError:
        return (component, None)


def parse_defines(
    root: xml.etree.ElementTree.Element,
    component_id: str,
    device_core: str | None = None,
) -> list[str]:
    """Parse pre-processor definitions for a component.

    Schema:
        <defines>
          <define name="EXAMPLE" value="1" device_cores="{device_core}..."/>
          <define name="OTHER" device_cores="{device_core}..."/>
        </defines>

    Args:
        root: root of element tree.
        component_id: id of component to return.
        device_core: name of core to filter sources for.

    Returns:
        list of str NAME=VALUE or NAME for the component.
    """
    xpath = f'./components/component[@id="{component_id}"]/defines/define'
    return list(
        _parse_define(define)
        for define in root.findall(xpath)
        if _element_is_compatible_with_device_core(define, device_core)
    )


def _parse_define(define: xml.etree.ElementTree.Element) -> str:
    """Parse <define> manifest stanza.

    Schema:
        <define name="EXAMPLE" value="1"/>
        <define name="OTHER"/>

    Args:
        define: XML Element for <define>.

    Returns:
        str with a value NAME=VALUE or NAME.
    """
    name = define.attrib['name']
    value = define.attrib.get('value', None)
    if value is None:
        return name

    return f'{name}={value}'


def parse_include_paths(
    root: xml.etree.ElementTree.Element,
    component_id: str,
    device_core: str | None = None,
) -> list[pathlib.Path]:
    """Parse include directories for a component.

    Schema:
        <component id="{component_id}" package_base_path="component">
          <include_paths>
            <include_path relative_path="./" type="c_include"
                          device_cores="{device_core}..."/>
          </include_paths>
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.
        device_core: name of core to filter sources for.

    Returns:
        list of include directories for the component.
    """
    (component, base_path) = get_component(root, component_id)
    if component is None:
        return []

    include_paths: list[pathlib.Path] = []
    for include_type in ('c_include', 'asm_include'):
        include_xpath = f'./include_paths/include_path[@type="{include_type}"]'

        include_paths.extend(
            _parse_include_path(include_path, base_path)
            for include_path in component.findall(include_xpath)
            if _element_is_compatible_with_device_core(
                include_path, device_core
            )
        )
    return include_paths


def _parse_include_path(
    include_path: xml.etree.ElementTree.Element,
    base_path: pathlib.Path | None,
) -> pathlib.Path:
    """Parse <include_path> manifest stanza.

    Schema:
        <include_path relative_path="./" type="c_include"/>

    Args:
        include_path: XML Element for <input_path>.
        base_path: prefix for paths.

    Returns:
        Path, prefixed with `base_path`.
    """
    path = pathlib.Path(include_path.attrib['relative_path'])
    if base_path is None:
        return path
    return base_path / path


def parse_headers(
    root: xml.etree.ElementTree.Element,
    component_id: str,
    device_core: str | None = None,
) -> list[pathlib.Path]:
    """Parse header files for a component.

    Schema:
        <component id="{component_id}" package_base_path="component">
          <source relative_path="./" type="c_include"
                  device_cores="{device_core}...">
            <files mask="example.h"/>
          </source>
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.
        device_core: name of core to filter sources for.

    Returns:
        list of header files for the component.
    """
    return _parse_sources(
        root, component_id, 'c_include', device_core=device_core
    )


def parse_sources(
    root: xml.etree.ElementTree.Element,
    component_id: str,
    device_core: str | None = None,
) -> list[pathlib.Path]:
    """Parse source files for a component.

    Schema:
        <component id="{component_id}" package_base_path="component">
          <source relative_path="./" type="src" device_cores="{device_core}...">
            <files mask="example.cc"/>
          </source>
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.
        device_core: name of core to filter sources for.

    Returns:
        list of source files for the component.
    """
    source_files = []
    for source_type in ('src', 'src_c', 'src_cpp', 'asm_include'):
        source_files.extend(
            _parse_sources(
                root, component_id, source_type, device_core=device_core
            )
        )
    return source_files


def parse_libs(
    root: xml.etree.ElementTree.Element,
    component_id: str,
    device_core: str | None = None,
) -> list[pathlib.Path]:
    """Parse pre-compiled libraries for a component.

    Schema:
        <component id="{component_id}" package_base_path="component">
          <source relative_path="./" type="lib" device_cores="{device_core}...">
            <files mask="example.a"/>
          </source>
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.
        device_core: name of core to filter sources for.

    Returns:
        list of pre-compiler libraries for the component.
    """
    return _parse_sources(root, component_id, 'lib', device_core=device_core)


def _parse_sources(
    root: xml.etree.ElementTree.Element,
    component_id: str,
    source_type: str,
    device_core: str | None = None,
) -> list[pathlib.Path]:
    """Parse <source> manifest stanza.

    Schema:
        <component id="{component_id}" package_base_path="component">
          <source relative_path="./" type="{source_type}"
                  device_cores="{device_core}...">
            <files mask="example.h"/>
          </source>
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.
        source_type: type of source to search for.
        device_core: name of core to filter sources for.

    Returns:
        list of source files for the component.
    """
    (component, base_path) = get_component(root, component_id)
    if component is None:
        return []

    sources: list[pathlib.Path] = []
    source_xpath = f'./source[@type="{source_type}"]'
    for source in component.findall(source_xpath):
        if not _element_is_compatible_with_device_core(source, device_core):
            continue

        relative_path = pathlib.Path(source.attrib['relative_path'])
        if base_path is not None:
            relative_path = base_path / relative_path

        sources.extend(
            relative_path / files.attrib['mask']
            for files in source.findall('./files')
        )
    return sources


def parse_dependencies(
    root: xml.etree.ElementTree.Element, component_id: str
) -> list[str]:
    """Parse the list of dependencies for a component.

    Optional dependencies are ignored for parsing since they have to be
    included explicitly.

    Schema:
        <dependencies>
          <all>
            <component_dependency value="component"/>
            <component_dependency value="component"/>
            <any_of>
              <component_dependency value="component"/>
              <component_dependency value="component"/>
            </any_of>
          </all>
        </dependencies>

    Args:
        root: root of element tree.
        component_id: id of component to return.

    Returns:
        list of component id dependencies of the component.
    """
    dependencies = []
    xpath = f'./components/component[@id="{component_id}"]/dependencies/*'
    for dependency in root.findall(xpath):
        dependencies.extend(_parse_dependency(dependency))
    return dependencies


def _parse_dependency(dependency: xml.etree.ElementTree.Element) -> list[str]:
    """Parse <all>, <any_of>, and <component_dependency> manifest stanzas.

    Schema:
        <all>
          <component_dependency value="component"/>
          <component_dependency value="component"/>
          <any_of>
            <component_dependency value="component"/>
            <component_dependency value="component"/>
          </any_of>
        </all>

    Args:
        dependency: XML Element of dependency.

    Returns:
        list of component id dependencies.
    """
    if dependency.tag == 'component_dependency':
        return [dependency.attrib['value']]
    if dependency.tag == 'all':
        dependencies = []
        for subdependency in dependency:
            dependencies.extend(_parse_dependency(subdependency))
        return dependencies
    if dependency.tag == 'any_of':
        # Explicitly ignore.
        return []

    # Unknown dependency tag type.
    return []


def check_dependencies(
    root: xml.etree.ElementTree.Element,
    component_id: str,
    include: Collection[str],
    exclude: Container[str] | None = None,
    device_core: str | None = None,
) -> bool:
    """Check the list of optional dependencies for a component.

    Verifies that the optional dependencies for a component are satisfied by
    components listed in `include` or `exclude`.

    Args:
        root: root of element tree.
        component_id: id of component to check.
        include: collection of component ids included in the project.
        exclude: optional container of component ids explicitly excluded from
            the project.
        device_core: name of core to filter sources for.

    Returns:
        True if dependencies are satisfied, False if not.
    """
    xpath = f'./components/component[@id="{component_id}"]/dependencies/*'
    for dependency in root.findall(xpath):
        if not _check_dependency(
            dependency, include, exclude=exclude, device_core=device_core
        ):
            return False
    return True


def _check_dependency(
    dependency: xml.etree.ElementTree.Element,
    include: Collection[str],
    exclude: Container[str] | None = None,
    device_core: str | None = None,
) -> bool:
    """Check a dependency for a component.

    Verifies that the given {dependency} is satisfied by components listed in
    `include` or `exclude`.

    Args:
        dependency: XML Element of dependency.
        include: collection of component ids included in the project.
        exclude: optional container of component ids explicitly excluded from
            the project.
        device_core: name of core to filter sources for.

    Returns:
        True if dependencies are satisfied, False if not.
    """
    if dependency.tag == 'component_dependency':
        component_id = dependency.attrib['value']
        return component_id in include or (
            exclude is not None and component_id in exclude
        )
    if dependency.tag == 'all':
        for subdependency in dependency:
            if not _check_dependency(
                subdependency, include, exclude=exclude, device_core=device_core
            ):
                return False
        return True
    if dependency.tag == 'any_of':
        for subdependency in dependency:
            if _check_dependency(
                subdependency, include, exclude=exclude, device_core=device_core
            ):
                return True

        tree = xml.etree.ElementTree.tostring(dependency).decode('utf-8')
        print(f'Unsatisfied dependency from: {tree}', file=sys.stderr)
        return False

    # Unknown dependency tag type.
    return True


def create_project(
    root: xml.etree.ElementTree.Element,
    include: Collection[str],
    exclude: Container[str] | None = None,
    device_core: str | None = None,
) -> tuple[
    list[str],
    list[str],
    list[pathlib.Path],
    list[pathlib.Path],
    list[pathlib.Path],
    list[pathlib.Path],
]:
    """Create a project from a list of specified components.

    Args:
        root: root of element tree.
        include: collection of component ids included in the project.
        exclude: container of component ids excluded from the project.
        device_core: name of core to filter sources for.

    Returns:
        (component_ids, defines, include_paths, headers, sources, libs) for the
        project.
    """
    # Build the project list from the list of included components by expanding
    # dependencies.
    project_list = []
    pending_list = list(include)
    while len(pending_list) > 0:
        component_id = pending_list.pop(0)
        if component_id in project_list:
            continue
        if exclude is not None and component_id in exclude:
            continue

        project_list.append(component_id)
        pending_list.extend(parse_dependencies(root, component_id))

    return (
        project_list,
        sum(
            (
                parse_defines(root, component_id, device_core=device_core)
                for component_id in project_list
            ),
            [],
        ),
        sum(
            (
                parse_include_paths(root, component_id, device_core=device_core)
                for component_id in project_list
            ),
            [],
        ),
        sum(
            (
                parse_headers(root, component_id, device_core=device_core)
                for component_id in project_list
            ),
            [],
        ),
        sum(
            (
                parse_sources(root, component_id, device_core=device_core)
                for component_id in project_list
            ),
            [],
        ),
        sum(
            (
                parse_libs(root, component_id, device_core=device_core)
                for component_id in project_list
            ),
            [],
        ),
    )


class Project:
    """Self-contained MCUXpresso project.

    Properties:
        component_ids: list of component ids compromising the project.
        defines: list of compiler definitions to build the project.
        include_dirs: list of include directory paths needed for the project.
        headers: list of header paths exported by the project.
        sources: list of source file paths built as part of the project.
        libs: list of libraries linked to the project.
        dependencies_satisfied: True if the project dependencies are satisfied.
    """

    @classmethod
    def from_file(
        cls,
        manifest_path: pathlib.Path,
        include: Collection[str],
        exclude: Container[str] | None = None,
        device_core: str | None = None,
    ):
        """Create a self-contained project with the specified components.

        Args:
            manifest_path: path to SDK manifest XML.
            include: collection of component ids included in the project.
            exclude: container of component ids excluded from the project.
            device_core: name of core to filter sources for.
        """
        tree = xml.etree.ElementTree.parse(manifest_path)
        root = tree.getroot()
        return cls(
            root, include=include, exclude=exclude, device_core=device_core
        )

    def __init__(
        self,
        manifest: xml.etree.ElementTree.Element,
        include: Collection[str],
        exclude: Container[str] | None = None,
        device_core: str | None = None,
    ):
        """Create a self-contained project with the specified components.

        Args:
            manifest: parsed manifest XML.
            include: collection of component ids included in the project.
            exclude: container of component ids excluded from the project.
            device_core: name of core to filter sources for.
        """
        (
            self.component_ids,
            self.defines,
            self.include_dirs,
            self.headers,
            self.sources,
            self.libs,
        ) = create_project(
            manifest, include, exclude=exclude, device_core=device_core
        )

        for component_id in self.component_ids:
            if not check_dependencies(
                manifest,
                component_id,
                self.component_ids,
                exclude=exclude,
                device_core=device_core,
            ):
                self.dependencies_satisfied = False
                return

        self.dependencies_satisfied = True



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_system_protos/device_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%pw_system_protos/device_service.proto\x12\x0fpw.system.proto\"\x0f\n\rRebootRequest\"\x10\n\x0eRebootResponse\"\x0e\n\x0c\x43rashRequest\"\x0f\n\rCrashResponse2\xa6\x01\n\rDeviceService\x12K\n\x06Reboot\x12\x1e.pw.system.proto.RebootRequest\x1a\x1f.pw.system.proto.RebootResponse\"\x00\x12H\n\x05\x43rash\x12\x1d.pw.system.proto.CrashRequest\x1a\x1e.pw.system.proto.CrashResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_system_protos.device_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_REBOOTREQUEST']._serialized_start=58
  _globals['_REBOOTREQUEST']._serialized_end=73
  _globals['_REBOOTRESPONSE']._serialized_start=75
  _globals['_REBOOTRESPONSE']._serialized_end=91
  _globals['_CRASHREQUEST']._serialized_start=93
  _globals['_CRASHREQUEST']._serialized_end=107
  _globals['_CRASHRESPONSE']._serialized_start=109
  _globals['_CRASHRESPONSE']._serialized_end=124
  _globals['_DEVICESERVICE']._serialized_start=127
  _globals['_DEVICESERVICE']._serialized_end=293
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/bundled_update.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from pw_protobuf_protos import common_pb2 as pw__protobuf__protos_dot_common__pb2
from pw_tokenizer_proto import options_pb2 as pw__tokenizer__proto_dot_options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'pw_software_update/bundled_update.proto\x12\x12pw.software_update\x1a\x19google/protobuf/any.proto\x1a\x1fpw_protobuf_protos/common.proto\x1a pw_tokenizer_proto/options.proto\"\xa2\x01\n\x12\x42undledUpdateState\"\x8b\x01\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08INACTIVE\x10\x01\x12\x10\n\x0cTRANSFERRING\x10\x02\x12\x0f\n\x0bTRANSFERRED\x10\x03\x12\r\n\tVERIFYING\x10\x04\x12\x0c\n\x08VERIFIED\x10\x05\x12\x0c\n\x08\x41PPLYING\x10\x06\x12\x0c\n\x08\x41\x42ORTING\x10\x07\x12\x0c\n\x08\x46INISHED\x10\x08\"\x91\x01\n\x13\x42undledUpdateResult\"z\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07SUCCESS\x10\x01\x12\x11\n\rUNKNOWN_ERROR\x10\x02\x12\x0b\n\x07\x41\x42ORTED\x10\x03\x12\x13\n\x0fTRANSFER_FAILED\x10\x04\x12\x11\n\rVERIFY_FAILED\x10\x05\x12\x10\n\x0c\x41PPLY_FAILED\x10\x06\"\xc7\x03\n\x13\x42undledUpdateStatus\x12:\n\x05state\x18\x01 \x01(\x0e\x32+.pw.software_update.BundledUpdateState.Enum\x12\x41\n\x06result\x18\x02 \x01(\x0e\x32,.pw.software_update.BundledUpdateResult.EnumH\x00\x88\x01\x01\x12\x34\n\'current_state_progress_hundreth_percent\x18\x03 \x01(\rH\x01\x88\x01\x01\x12\x18\n\x0btransfer_id\x18\x04 \x01(\rH\x02\x88\x01\x01\x12\x1c\n\x0f\x62undle_filename\x18\x05 \x01(\tH\x03\x88\x01\x01\x12\x17\n\x04note\x18\x06 \x01(\x0c\x42\x04\x80\xaf&\x00H\x04\x88\x01\x01\x12\x32\n\x0f\x65xtended_status\x18\x07 \x01(\x0b\x32\x14.google.protobuf.AnyH\x05\x88\x01\x01\x42\t\n\x07_resultB*\n(_current_state_progress_hundreth_percentB\x0e\n\x0c_transfer_idB\x12\n\x10_bundle_filenameB\x07\n\x05_noteB\x12\n\x10_extended_status\"@\n\x0cStartRequest\x12\x1c\n\x0f\x62undle_filename\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x12\n\x10_bundle_filename2\x95\x04\n\rBundledUpdate\x12H\n\tGetStatus\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12R\n\x05Start\x12 .pw.software_update.StartRequest\x1a\'.pw.software_update.BundledUpdateStatus\x12M\n\x0eSetTransferred\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x45\n\x06Verify\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05\x41pply\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05\x41\x62ort\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05Reset\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatusB\"\n\x1e\x64\x65v.pigweed.pw_software_updateP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_software_update.bundled_update_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\036dev.pigweed.pw_software_updateP\001'
  _globals['_BUNDLEDUPDATESTATUS'].fields_by_name['note']._options = None
  _globals['_BUNDLEDUPDATESTATUS'].fields_by_name['note']._serialized_options = b'\200\257&\000'
  _globals['_BUNDLEDUPDATESTATE']._serialized_start=158
  _globals['_BUNDLEDUPDATESTATE']._serialized_end=320
  _globals['_BUNDLEDUPDATESTATE_ENUM']._serialized_start=181
  _globals['_BUNDLEDUPDATESTATE_ENUM']._serialized_end=320
  _globals['_BUNDLEDUPDATERESULT']._serialized_start=323
  _globals['_BUNDLEDUPDATERESULT']._serialized_end=468
  _globals['_BUNDLEDUPDATERESULT_ENUM']._serialized_start=346
  _globals['_BUNDLEDUPDATERESULT_ENUM']._serialized_end=468
  _globals['_BUNDLEDUPDATESTATUS']._serialized_start=471
  _globals['_BUNDLEDUPDATESTATUS']._serialized_end=926
  _globals['_STARTREQUEST']._serialized_start=928
  _globals['_STARTREQUEST']._serialized_end=992
  _globals['_BUNDLEDUPDATE']._serialized_start=995
  _globals['_BUNDLEDUPDATE']._serialized_end=1528
# @@protoc_insertion_point(module_scope)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Decodes the in-memory representation of a sized-entry ring buffer."""

import struct
from typing import Iterable

_HEADER = struct.Struct('III')  # data_size_bytes, head, tail


def _decode_leb128(
    data: bytes, offset: int = 0, max_bits: int = 32
) -> tuple[int, int]:
    count = value = shift = 0

    while offset < len(data):
        byte = data[offset]

        count += 1
        value |= (byte & 0x7F) << shift

        if not byte & 0x80:
            return offset + count, value

        shift += 7
        if shift >= max_bits:
            raise ValueError(f'Varint exceeded {max_bits}-bit limit')

    raise ValueError(f'Unterminated varint {data[offset:]!r}')


def parse(queue: bytes) -> Iterable[bytes]:
    """Decodes the in-memory representation of a variable-length entry queue.

    Args:
      queue: The bytes representation of a variable-length entry queue.

    Yields:
      Each entry in the buffer as bytes.
    """
    array_size_bytes, head, tail = _HEADER.unpack_from(queue)

    total_encoded_size = _HEADER.size + array_size_bytes
    if len(queue) < total_encoded_size:
        raise ValueError(
            f'Ring buffer data ({len(queue)} B) is smaller than the encoded '
            f'size ({total_encoded_size} B)'
        )

    data = queue[_HEADER.size : total_encoded_size]

    if tail < head:
        data = data[head:] + data[:tail]
    else:
        data = data[head:tail]

    index = 0
    while index < len(data):
        index, size = _decode_leb128(data, index)

        if index + size > len(data):
            raise ValueError(
                f'Corruption detected; '
                f'encoded size {size} B is too large for a {len(data)} B array'
            )
        yield data[index : index + size]

        index += size



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Device tracing classes to interact with targets via RPC."""

import os
import logging
import tempfile

from pw_rpc.callback_client.errors import RpcError
from pw_system.device import Device
from pw_trace import trace
from pw_trace_tokenized import trace_tokenized

_LOG = logging.getLogger(__package__)
DEFAULT_TICKS_PER_SECOND = 1000


class DeviceWithTracing(Device):
    """Represents an RPC Client for a device running a Pigweed target with
    tracing.

    The target must have RPC support for the following services:
     - tracing

    Note: use this class as a base for specialized device representations.
    """

    def __init__(
        self,
        *device_args,
        ticks_per_second: int | None = None,
        time_offset: int = 0,
        **device_kwargs,
    ):
        super().__init__(*device_args, **device_kwargs)

        self.time_offset = time_offset

        if ticks_per_second:
            self.ticks_per_second = ticks_per_second
        else:
            self.ticks_per_second = self.get_ticks_per_second()
        _LOG.info('ticks_per_second set to %i', self.ticks_per_second)

    def get_ticks_per_second(self) -> int:
        trace_service = self.rpcs.pw.trace.proto.TraceService
        try:
            resp = trace_service.GetClockParameters()
            if not resp.status.ok():
                _LOG.error(
                    'Failed to get clock parameters: %s. Using default value',
                    resp.status,
                )
                return DEFAULT_TICKS_PER_SECOND
        except RpcError as rpc_err:
            _LOG.exception('%s. Using default value', rpc_err)
            return DEFAULT_TICKS_PER_SECOND

        return resp.response.clock_parameters.tick_period_seconds_denominator

    def start_tracing(self) -> None:
        """Turns on tracing on this device."""
        trace_service = self.rpcs.pw.trace.proto.TraceService
        trace_service.Start()

    def stop_tracing(self, trace_output_path: str = "trace.json") -> None:
        """Turns off tracing on this device and downloads the trace file."""
        trace_service = self.rpcs.pw.trace.proto.TraceService
        resp = trace_service.Stop()

        # If there's no tokenizer, there's no need to transfer the trace
        # file from the device after stopping tracing, as there's not much
        # that can be done with it.
        if not self.detokenizer:
            _LOG.error('No tokenizer specified. Not transfering trace')
            return

        trace_bin_path = tempfile.NamedTemporaryFile(delete=False)
        trace_bin_path.close()
        try:
            if not self.transfer_file(
                resp.response.file_id, trace_bin_path.name
            ):
                return

            with open(trace_bin_path.name, 'rb') as bin_file:
                trace_data = bin_file.read()
                events = trace_tokenized.get_trace_events(
                    [self.detokenizer.database],
                    trace_data,
                    self.ticks_per_second,
                    self.time_offset,
                )
                json_lines = trace.generate_trace_json(events)
                trace_tokenized.save_trace_file(json_lines, trace_output_path)

            _LOG.info(
                'Wrote trace file %s',
                trace_output_path,
            )
        finally:
            os.remove(trace_bin_path.name)



#!/usr/bin/env python3
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Arduino Core Installer."""

import argparse
import importlib.resources
import logging
import os
from pathlib import Path
import platform
import shutil

try:
    from pw_arduino_build import file_operations

except ImportError:
    # Load this directory if pw_arduino_build is not available.
    import file_operations  # type: ignore

_LOG = logging.getLogger(__name__)


class ArduinoCoreNotSupported(Exception):
    """Exception raised when a given core can not be installed."""


class ArduinoCoreInstallationFailed(Exception):
    """Exception raised when a given core failed to be installed."""


_ARDUINO_CORE_ARTIFACTS: dict[str, dict] = {
    # pylint: disable=line-too-long
    "teensy": {
        "all": {
            "core": {
                "version": "1.58.1",
                "url": "https://www.pjrc.com/teensy/td_158-1/teensy-package.tar.bz2",
                "file_name": "teensy-package.tar.bz2",
                "sha256": "3a3f3728045621d25068c5b5dfc24bf171550127e9fae4d0e8be2574c6636cff",
            }
        },
        "Linux": {
            "teensy-tools": {
                "url": "https://www.pjrc.com/teensy/td_158/teensy-tools-linux64.tar.bz2",
                "file_name": "teensy-tools-linux64.tar.bz2",
                "sha256": "0a272575ca42b4532ab89516df160e1d68e449fe1538c0bd71dbb768f1b3c0b6",
                "version": "1.58.0",
            },
        },
        # TODO(tonymd): Handle 32-bit Linux Install?
        "Linux32": {
            "teensy-tools": {
                "url": "https://www.pjrc.com/teensy/td_158/teensy-tools-linux32.tar.bz2",
                "file_name": "teensy-tools-linux32.tar.bz2",
                "sha256": "995d974935c8118ad6d4c191206453dd8f57c1e299264bb4cffcc62c96c6d077",
                "version": "1.58.0",
            },
        },
        # TODO(tonymd): Handle ARM32 (Raspberry Pi) Install?
        "LinuxARM32": {
            "teensy-tools": {
                "url": "https://www.pjrc.com/teensy/td_158/teensy-tools-linuxarm.tar.bz2",
                "file_name": "teensy-tools-linuxarm.tar.bz2",
                "sha256": "88cf8e55549f5d5937fa7dbc763cad49bd3680d4e5185b318c667f541035e633",
                "version": "1.58.0",
            },
        },
        # TODO(tonymd): Handle ARM64 Install?
        "LinuxARM64": {
            "teensy-tools": {
                "url": "https://www.pjrc.com/teensy/td_158/teensy-tools-linuxaarch64.tar.bz2",
                "file_name": "teensy-tools-linuxaarch64.tar.bz2",
                "sha256": "a20b1c5e91fe51c3b6591e4cfcf711d4a4c0a0bb5120c59d1c8dd8d32ae44e31",
                "version": "1.58.0",
            },
        },
        "Darwin": {
            "teensy-tools": {
                "url": "https://www.pjrc.com/teensy/td_158/teensy-tools-macos.tar.bz2",
                "file_name": "teensy-tools-macos.tar.bz2",
                "sha256": "d386412e38fe6dd6c5d849c2b1f8eea00cbf7bc3659fb6ba9f83cebfb736924b",
                "version": "1.58.0",
            },
        },
        "Windows": {
            "teensy-tools": {
                "url": "https://www.pjrc.com/teensy/td_158/teensy-tools-windows.tar.bz2",
                "file_name": "teensy-tools-windows.tar.bz2",
                "sha256": "206315ddc82381d2da92da9f633a1719e00c0e8f5432acfed434573409a48de1",
                "version": "1.58.0",
            },
        },
    },
    "adafruit-samd": {
        "all": {
            "core": {
                "version": "1.6.2",
                "url": "https://github.com/adafruit/ArduinoCore-samd/archive/1.6.2.tar.gz",
                "file_name": "adafruit-samd-1.6.2.tar.gz",
                "sha256": "5875f5bc05904c10e6313f02653f28f2f716db639d3d43f5a1d8a83d15339d64",
            }
        },
        "Linux": {},
        "Darwin": {},
        "Windows": {},
    },
    "arduino-samd": {
        "all": {
            "core": {
                "version": "1.8.8",
                "url": "http://downloads.arduino.cc/cores/samd-1.8.8.tar.bz2",
                "file_name": "arduino-samd-1.8.8.tar.bz2",
                "sha256": "7b93eb705cba9125d9ee52eba09b51fb5fe34520ada351508f4253abbc9f27fa",
            }
        },
        "Linux": {},
        "Darwin": {},
        "Windows": {},
    },
    "stm32duino": {
        "all": {
            "core": {
                "version": "2.6.0",
                "url": "https://github.com/stm32duino/Arduino_Core_STM32/archive/2.6.0.tar.gz",
                "file_name": "stm32duino-2.6.0.tar.gz",
                "sha256": "53f37df1202b1bccfb353e4775200f63b36d487fe734fdb8ca9bfa00c2f3429f",
            }
        },
        "Linux": {},
        "Darwin": {},
        "Windows": {},
    },
}


def install_core_command(args: argparse.Namespace):
    install_core(args.prefix, args.core_name)


def install_core(prefix, core_name):
    install_prefix = os.path.realpath(
        os.path.expanduser(os.path.expandvars(prefix))
    )
    install_dir = os.path.join(install_prefix, core_name)
    cache_dir = os.path.join(install_prefix, ".cache", core_name)

    if core_name in supported_cores():
        shutil.rmtree(install_dir, ignore_errors=True)
        os.makedirs(install_dir, exist_ok=True)
        os.makedirs(cache_dir, exist_ok=True)

    if core_name == "teensy":
        install_teensy_core(install_prefix, install_dir, cache_dir)
        apply_teensy_patches(install_dir)
    elif core_name == "adafruit-samd":
        install_adafruit_samd_core(install_prefix, install_dir, cache_dir)
    elif core_name == "stm32duino":
        install_stm32duino_core(install_prefix, install_dir, cache_dir)
    elif core_name == "arduino-samd":
        install_arduino_samd_core(install_prefix, install_dir, cache_dir)
    else:
        raise ArduinoCoreNotSupported(
            "Invalid core '{}'. Supported cores: {}".format(
                core_name, ", ".join(supported_cores())
            )
        )


def supported_cores():
    return _ARDUINO_CORE_ARTIFACTS.keys()


def install_teensy_core(_install_prefix: str, install_dir: str, cache_dir: str):
    """Install teensy core files and tools."""
    # Install the Teensy core source files
    artifacts = _ARDUINO_CORE_ARTIFACTS["teensy"]["all"]["core"]
    core_tarfile = file_operations.download_to_cache(
        url=artifacts["url"],
        expected_sha256sum=artifacts["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=artifacts["file_name"],
    )

    package_path = os.path.join(
        install_dir, "hardware", "avr", artifacts["version"]
    )
    os.makedirs(package_path, exist_ok=True)
    file_operations.extract_archive(core_tarfile, package_path, cache_dir)

    expected_files = [
        Path(package_path) / 'boards.txt',
        Path(package_path) / 'platform.txt',
    ]

    if any(not expected_file.is_file() for expected_file in expected_files):
        expected_files_str = "".join(
            list(f"  {expected_file}\n" for expected_file in expected_files)
        )

        raise ArduinoCoreInstallationFailed(
            "\n\nError: Installation Failed.\n"
            "Please remove the package:\n\n"
            "  pw package remove teensy\n\n"
            "Try again and ensure the following files exist:\n\n"
            + expected_files_str
        )

    teensy_tools = _ARDUINO_CORE_ARTIFACTS["teensy"][platform.system()]

    for tool_name, artifacts in teensy_tools.items():
        tool_tarfile = file_operations.download_to_cache(
            url=artifacts["url"],
            expected_sha256sum=artifacts["sha256"],
            cache_directory=cache_dir,
            downloaded_file_name=artifacts["file_name"],
        )

        tool_path = os.path.join(
            install_dir, "tools", tool_name, artifacts["version"]
        )

        os.makedirs(tool_path, exist_ok=True)
        file_operations.extract_archive(tool_tarfile, tool_path, cache_dir)

    return True


def apply_teensy_patches(install_dir):
    # On Mac the "hardware" directory is a symlink:
    #   ls -l third_party/arduino/cores/teensy/
    #   hardware -> Teensyduino.app/Contents/Java/hardware
    # Resolve paths since `git apply` doesn't work if a path is beyond a
    # symbolic link.
    patch_root_path = (
        Path(install_dir) / "hardware/avr/1.58.1/cores"
    ).resolve()

    # Get all *.diff files for the teensy core.
    patches_python_package = 'pw_arduino_build.core_patches.teensy'

    patch_file_names = sorted(
        patch
        for patch in importlib.resources.contents(patches_python_package)
        if Path(patch).suffix in ['.diff']
    )

    # Apply each patch file.
    for diff_name in patch_file_names:
        with importlib.resources.path(
            patches_python_package, diff_name
        ) as diff_path:
            file_operations.git_apply_patch(
                patch_root_path.as_posix(),
                diff_path.as_posix(),
                unsafe_paths=True,
            )


def install_arduino_samd_core(
    install_prefix: str, install_dir: str, cache_dir: str
):
    artifacts = _ARDUINO_CORE_ARTIFACTS["arduino-samd"]["all"]["core"]
    core_tarfile = file_operations.download_to_cache(
        url=artifacts["url"],
        expected_sha256sum=artifacts["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=artifacts["file_name"],
    )

    package_path = os.path.join(
        install_dir, "hardware", "samd", artifacts["version"]
    )
    os.makedirs(package_path, exist_ok=True)
    file_operations.extract_archive(core_tarfile, package_path, cache_dir)
    original_working_dir = os.getcwd()
    os.chdir(install_prefix)
    # TODO(tonymd): Fetch core/tools as specified by:
    # http://downloads.arduino.cc/packages/package_index.json
    os.chdir(original_working_dir)
    return True


def install_adafruit_samd_core(
    install_prefix: str, install_dir: str, cache_dir: str
):
    artifacts = _ARDUINO_CORE_ARTIFACTS["adafruit-samd"]["all"]["core"]
    core_tarfile = file_operations.download_to_cache(
        url=artifacts["url"],
        expected_sha256sum=artifacts["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=artifacts["file_name"],
    )

    package_path = os.path.join(
        install_dir, "hardware", "samd", artifacts["version"]
    )
    os.makedirs(package_path, exist_ok=True)
    file_operations.extract_archive(core_tarfile, package_path, cache_dir)

    original_working_dir = os.getcwd()
    os.chdir(install_prefix)
    # TODO(tonymd): Fetch platform specific tools as specified by:
    # https://adafruit.github.io/arduino-board-index/package_adafruit_index.json
    # Specifically:
    #   https://github.com/ARM-software/CMSIS_5/archive/5.4.0.tar.gz
    os.chdir(original_working_dir)
    return True


def install_stm32duino_core(install_prefix, install_dir, cache_dir):
    artifacts = _ARDUINO_CORE_ARTIFACTS["stm32duino"]["all"]["core"]
    core_tarfile = file_operations.download_to_cache(
        url=artifacts["url"],
        expected_sha256sum=artifacts["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=artifacts["file_name"],
    )

    package_path = os.path.join(
        install_dir, "hardware", "stm32", artifacts["version"]
    )
    os.makedirs(package_path, exist_ok=True)
    file_operations.extract_archive(core_tarfile, package_path, cache_dir)
    original_working_dir = os.getcwd()
    os.chdir(install_prefix)
    # TODO(tonymd): Fetch platform specific tools as specified by:
    # https://github.com/stm32duino/BoardManagerFiles/raw/HEAD/STM32/package_stm_index.json
    os.chdir(original_working_dir)
    return True



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Install and check status of picotool."""

import logging
import os
from pathlib import Path
import shutil
import subprocess
from typing import Sequence

import pw_package.git_repo
import pw_package.package_manager

_LOG = logging.getLogger(__package__)


def force_copy(source: Path, destination: Path):
    _LOG.info('Copy %s -> %s', source, destination)
    # ensure the destination directory exists
    # otherwise the copy will fail on mac.
    dirname = os.path.dirname(destination)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)
    destination.unlink(missing_ok=True)
    shutil.copy(source, destination)


class Picotool(pw_package.package_manager.Package):
    """Install and check status of picotool."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, name='picotool', **kwargs)

    def install(self, path: Path) -> None:
        env = os.environ.copy()

        def log_and_run(command: Sequence[str], **kwargs):
            _LOG.info('==> %s', ' '.join(command))
            return subprocess.run(
                command,
                env=env,
                check=True,
                **kwargs,
            )

        log_and_run(('bazel', 'build', '@picotool'))
        build_path = log_and_run(
            ('bazel', 'cquery', '@picotool', '--output=files'),
            capture_output=True,
            text=True,
        ).stdout.strip()

        picotool_bin = path / 'out' / 'picotool'
        force_copy(build_path, picotool_bin)

        _LOG.info('Done! picotool binary located at:')
        _LOG.info(picotool_bin)

        bootstrap_env_path = Path(env.get('_PW_ACTUAL_ENVIRONMENT_ROOT', ''))
        if bootstrap_env_path.is_dir() and picotool_bin.is_file():
            bin_path = (
                bootstrap_env_path / 'cipd' / 'packages' / 'pigweed' / 'bin'
            )
            destination_path = bin_path / picotool_bin.name
            force_copy(build_path, destination_path)

    def info(self, path: Path) -> Sequence[str]:
        return (f'{self.name} installed in: {path}',)


pw_package.package_manager.register(Picotool)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Download all Python packages required for a pw_python_venv."""

import argparse
import logging
from pathlib import Path
import shlex
import subprocess
import sys


_LOG = logging.getLogger('pw_build.generate_python_wheel_cache')


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--pip-download-log',
        type=Path,
        required=True,
        help='Path to the root gn build dir.',
    )
    parser.add_argument(
        '--wheel-dir',
        type=Path,
        required=True,
        help='Path to save wheel files.',
    )
    parser.add_argument(
        '--download-all-platforms',
        action='store_true',
        help='Download Python precompiled wheels for all platforms.',
    )
    parser.add_argument(
        '-r',
        '--requirement',
        type=Path,
        nargs='+',
        required=True,
        help='Requirements files',
    )
    return parser.parse_args()


def main(
    pip_download_log: Path,
    wheel_dir: Path,
    requirement: list[Path],
    download_all_platforms: bool = False,
) -> int:
    """Download all Python packages required for a pw_python_venv."""

    # Delete existing wheels from the out dir, there may be stale versions.
    # shutil.rmtree(wheel_dir, ignore_errors=True)
    wheel_dir.mkdir(parents=True, exist_ok=True)

    pip_download_args = [
        sys.executable,
        "-m",
        "pip",
        "--log",
        str(pip_download_log),
        "download",
        "--dest",
        str(wheel_dir),
    ]
    for req in requirement:
        pip_download_args.extend(["--requirement", str(req)])

    if not download_all_platforms:
        # Download for the current platform only.
        quoted_pip_download_args = ' '.join(
            shlex.quote(arg) for arg in pip_download_args
        )
        _LOG.info('Run ==> %s', quoted_pip_download_args)
        # Download packages
        subprocess.run(pip_download_args, check=True)
        return 0

    # DOCSTAG: [wheel-platform-args]
    # These platform args are derived from the cffi pypi package:
    #   https://pypi.org/project/cffi/#files
    # See also these pages on Python wheel filename format:
    #   https://peps.python.org/pep-0491/#file-name-convention
    # and Platform compatibility tags:
    #   https://packaging.python.org/en/latest/specifications/
    #      platform-compatibility-tags/
    platform_args = [
        '--platform=any',
        '--platform=macosx_10_9_universal2',
        '--platform=macosx_10_9_x86_64',
        '--platform=macosx_11_0_arm64',
        '--platform=manylinux2010_x86_64',
        '--platform=manylinux2014_aarch64',
        '--platform=manylinux2014_x86_64',
        '--platform=manylinux_2_17_aarch64',
        '--platform=manylinux_2_17_x86_64',
        '--platform=musllinux_1_1_x86_64',
        '--platform=win_amd64',
        # Note: These 32bit platforms are omitted
        # '--platform=manylinux2010_i686',
        # '--platform=manylinux2014_i686',
        # '--platform=manylinux_2_12_i686'
        # '--platform=musllinux_1_1_i686',
        # '--platform=win32',
    ]

    # Pigweed supports Python 3.8 and up.
    python_version_args = [
        [
            '--implementation=py3',
            '--abi=none',
        ],
        [
            '--implementation=cp',
            '--python-version=3.8',
            '--abi=cp38',
        ],
        [
            '--implementation=cp',
            '--python-version=3.9',
            '--abi=cp39',
        ],
        [
            '--implementation=cp',
            '--python-version=3.10',
            '--abi=cp310',
        ],
        [
            '--implementation=cp',
            '--python-version=3.11',
            '--abi=cp311',
        ],
    ]
    # DOCSTAG: [wheel-platform-args]

    # --no-deps is required when downloading binary packages for different
    # platforms other than the current one. The requirements.txt files already
    # has the fully expanded deps list using pip-compile so this is not a
    # problem.
    pip_download_args.append('--no-deps')

    # Run pip download once for each Python version. Multiple platform args can
    # be added to the same command.
    for py_version_args in python_version_args:
        for platform_arg in platform_args:
            final_pip_download_args = list(pip_download_args)
            final_pip_download_args.append(platform_arg)
            final_pip_download_args.extend(py_version_args)
            quoted_pip_download_args = ' '.join(
                shlex.quote(arg) for arg in final_pip_download_args
            )
            _LOG.info('')
            _LOG.info('Fetching packages for:')
            _LOG.info(
                'Python %s and Platforms: %s', py_version_args, platform_args
            )
            _LOG.info('Run ==> %s', quoted_pip_download_args)

            # Download packages
            subprocess.run(final_pip_download_args, check=True)

    return 0


if __name__ == '__main__':
    logging.basicConfig(format='%(message)s', level=logging.DEBUG)
    sys.exit(main(**vars(_parse_args())))



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Utilities for manipulating GN targets."""

import re

from typing import Iterable


class GnTarget:
    """Represents a Pigweed Build target.

    Attributes:
        attrs:  Map of GN variable names to lists of values. Recognized GN list
                variables include 'public', 'sources', 'inputs', 'deps', etc.
        origin: Reference to the data used to generate this target, e.g. a
                Bazel label.
    """

    def __init__(self, target_type: str, target_name: str) -> None:
        """Creates a GN target.

        Args:
            target_type: Type of the GN target, like 'pw_source_set'.
            target_name: Name of the GN target.
        """
        self.attrs: dict[str, list[str]] = {}
        self.origin: str | None = None
        self._type = target_type
        self._name = target_name

    def type(self) -> str:
        """Returns this object's GN target_type."""
        return self._type

    def name(self) -> str:
        """Returns this object's GN target_name."""
        return self._name

    def build_args(self) -> Iterable[str]:
        """Returns a list of GN build arguments referenced in paths.

        Third party source code is referenced by Pigweed build files using build
        arguments, e.g. 'public = [ "$dir_pw_third_party_foo/bar/baz.h" ]'. This
        method returns any build arguments used to locate source, header or
        input files so that the BUILD.gn file can import the build argument
        definition.
        """
        build_arg_pat = re.compile(r'(\$dir_pw_third_party_[^/:]*)')
        for field in ['public', 'sources', 'inputs', 'include_dirs']:
            for source_path in self.attrs.get(field, []):
                match = re.search(build_arg_pat, source_path)
                if match:
                    yield match.group(1)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Class for describing file filter patterns."""

from pathlib import Path
import re
from typing import Pattern, Iterable, Sequence


class FileFilter:
    """Allows checking if a path matches a series of filters.

    Positive filters (e.g. the file name matches a regex) and negative filters
    (path does not match a regular expression) may be applied.
    """

    def __init__(
        self,
        *,
        exclude: Iterable[Pattern | str] = (),
        endswith: Iterable[str] = (),
        name: Iterable[Pattern | str] = (),
        suffix: Iterable[str] = (),
    ) -> None:
        """Creates a FileFilter with the provided filters.

        Args:
            endswith: True if the end of the path is equal to any of the passed
                      strings
            exclude: If any of the passed regular expresion match return False.
                     This overrides and other matches.
            name: Regexs to match with file names(pathlib.Path.name). True if
                  the resulting regex matches the entire file name.
            suffix: True if final suffix (as determined by pathlib.Path) is
                    matched by any of the passed str.
        """
        self.exclude = tuple(re.compile(i) for i in exclude)

        self.endswith = tuple(endswith)
        self.name = tuple(re.compile(i) for i in name)
        self.suffix = tuple(suffix)

    def matches(self, path: str | Path) -> bool:
        """Returns true if the path matches any filter but not an exclude.

        If no positive filters are specified, any paths that do not match a
        negative filter are considered to match.

        If 'path' is a Path object it is rendered as a posix path (i.e.
        using "/" as the path seperator) before testing with 'exclude' and
        'endswith'.
        """

        posix_path = path.as_posix() if isinstance(path, Path) else path
        if any(bool(exp.search(posix_path)) for exp in self.exclude):
            return False

        # If there are no positive filters set, accept all paths.
        no_filters = not self.endswith and not self.name and not self.suffix

        path_obj = Path(path)
        return (
            no_filters
            or path_obj.suffix in self.suffix
            or any(regex.fullmatch(path_obj.name) for regex in self.name)
            or any(posix_path.endswith(end) for end in self.endswith)
        )

    def filter(self, paths: Sequence[str | Path]) -> Sequence[Path]:
        return [Path(x) for x in paths if self.matches(x)]



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Zephyr SDK pw_env_setup action plugin.

This action triggers `cmake -P <cache_path>/cmake/zephyr_sdk_export.cmake` after
CIPD setup.
"""
import os
import pathlib
import shutil
import subprocess


def run_action(env=None):
    """Project action to install the Zephyr SDK."""
    if not env:
        raise ValueError(f"Missing 'env', got %{env}")

    with env():
        cmake = shutil.which('cmake')

        zephyr_sdk = list(
            pathlib.Path(os.environ['PW_CIPD_CIPD_INSTALL_DIR']).glob(
                '**/cmake/zephyr_sdk_export.cmake'
            )
        )[0]
        subprocess.check_call(
            [cmake, '-P', str(zephyr_sdk)],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Parses the arguments in a Base64-encoded tokenized message.

This is useful for attempting to decode tokenized messages with arguments for
which the token is not recognized.
"""

from __future__ import annotations

import argparse
import base64
from dataclasses import dataclass
import logging
import sys
from typing import Collection, Iterable, Iterator, Sequence

import pw_cli.log
from pw_tokenizer.decode import FormatString, FormattedString

_LOG: logging.Logger = logging.getLogger('pw_tokenizer')

DEFAULT_FORMAT_SPECS = (
    '%s',
    '%d',
    '%f',
)

DEFAULT_MAX_ARGS = 8
PREFIX = '$'


def attempt_to_decode(
    arg_data: bytes,
    format_specs: Collection[str] = DEFAULT_FORMAT_SPECS,
    max_args: int = DEFAULT_MAX_ARGS,
    yield_failures: bool = False,
) -> Iterator[FormattedString]:
    """Attempts to decode arguments using the provided format specifiers."""
    format_strings = [(0, '')]  # (argument count, format string)

    # Each argument requires at least 1 byte.
    max_args = min(max_args, len(arg_data))

    while format_strings:
        arg_count, string = format_strings.pop(0)
        decode_attempt = FormatString(string).format(arg_data)

        if yield_failures or decode_attempt.ok():
            yield decode_attempt

        if arg_count < max_args:
            format_strings.extend(
                (arg_count + 1, string + spec) for spec in format_specs
            )


@dataclass(frozen=True)
class TokenizedMessage:
    string: str
    binary: bytes

    @property
    def token(self) -> int:
        return int.from_bytes(self.binary[:4], 'little')

    @property
    def binary_args(self) -> bytes:
        return self.binary[4:]

    @classmethod
    def parse(cls, message: str, prefix: str = '$') -> TokenizedMessage:
        if not message.startswith(prefix):
            raise ValueError(
                f'{message} does not start with {prefix!r} as expected'
            )

        binary = base64.b64decode(message[1:])

        if len(binary) < 4:
            raise ValueError(
                f'{message} is only {len(binary)} bytes; '
                'tokenized messages must be at least 4 bytes'
            )

        return cls(message, binary)


def _read_stdin():
    try:
        while True:
            yield input()
    except KeyboardInterrupt:
        return


def _text_list(items: Sequence, conjunction: str = 'or') -> str:
    if len(items) == 1:
        return str(items[0])

    return f'{", ".join(str(i) for i in items[:-1])} {conjunction} {items[-1]}'


def main(
    messages: Iterable[str],
    max_args: int,
    specs: Sequence[str],
    show_failures: bool,
) -> int:
    """Parses the arguments for a series of tokenized messages."""
    exit_code = 0

    for message in iter(messages) if messages else _read_stdin():
        if not message:
            continue

        if not message.startswith(PREFIX):
            message = PREFIX + message

        _LOG.info('Decoding arguments for %r', message)
        try:
            parsed = TokenizedMessage.parse(message)
        except ValueError as exc:
            _LOG.error('%s', exc)
            exit_code = 2
            continue

        _LOG.info(
            'Binary: %r [%s] (%d bytes)',
            parsed.binary,
            parsed.binary.hex(' ', 1),
            len(parsed.binary),
        )
        _LOG.info('Token:  0x%08x', parsed.token)
        _LOG.info(
            'Args:   %r [%s] (%d bytes)',
            parsed.binary_args,
            parsed.binary_args.hex(' ', 1),
            len(parsed.binary_args),
        )
        _LOG.info(
            'Decoding with up to %d %s arguments', max_args, _text_list(specs)
        )

        results = sorted(
            attempt_to_decode(
                parsed.binary_args, specs, max_args, show_failures
            ),
            key=FormattedString.score,
            reverse=True,
        )

        if not any(result.ok() for result in results):
            _LOG.warning(
                '  No combinations of up to %d %s arguments decoded '
                'successfully',
                max_args,
                _text_list(specs),
            )
            exit_code = 1

        for i, result in enumerate(results, 1):
            _LOG.info(  # pylint: disable=logging-fstring-interpolation
                f'  Attempt %{len(str(len(results)))}d: [%s] %s',
                i,
                ' '.join(str(a.specifier) for a in result.args),
                ' '.join(str(a) for a in result.args),
            )
        print()

    return exit_code


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        '--max-args',
        default=DEFAULT_MAX_ARGS,
        type=int,
        help='Maximum number of printf-style arguments',
    )
    parser.add_argument(
        '--specs',
        nargs='*',
        default=DEFAULT_FORMAT_SPECS,
        help='Which printf-style format specifiers to check',
    )
    parser.add_argument(
        '--show-failures',
        action='store_true',
        help='Show argument combintations that fail to decode',
    )
    parser.add_argument(
        'messages',
        nargs='*',
        help=(
            'Base64-encoded tokenized messages to decode; omit to read from '
            'stdin'
        ),
    )
    return parser.parse_args()


if __name__ == '__main__':
    pw_cli.log.install()
    sys.exit(main(**vars(_parse_args())))



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Utilities for handling singular/plural forms in logs and tooling."""


def plural(
    items_or_count,
    singular: str,
    count_format='',
    these: bool = False,
    number: bool = True,
    are: bool = False,
    exist: bool = False,
) -> str:
    """Returns the singular or plural form of a word based on a count.

    Args:
        items_or_count: Number of items or a collection of items
        singular: Singular form of the name of the item
        count_format: .format()-style specification for items_or_count
        these: Prefix the string with "this" or "these", depending on number
        number: Include the number in the return string (e.g., "3 things" vs.
            "things")
        are: Suffix the string with "is" or "are", depending on number
        exist: Suffix the string with "exists" or "exist", depending on number
    """

    try:
        count = len(items_or_count)
    except TypeError:
        count = items_or_count

    prefix = ('this ' if count == 1 else 'these ') if these else ''
    num = f'{count:{count_format}} ' if number else ''

    suffix = ''
    if are and exist:
        raise ValueError(f'cannot combine are ({are}) and exist ({exist})')
    if are:
        suffix = ' is' if count == 1 else ' are'
    if exist:
        suffix = ' exists' if count == 1 else ' exist'

    if singular.endswith('y'):
        result = f'{singular[:-1]}{"y" if count == 1 else "ies"}'
    elif singular.endswith('s'):
        result = f'{singular}{"" if count == 1 else "es"}'
    else:
        result = f'{singular}{"" if count == 1 else "s"}'

    return f'{prefix}{num}{result}{suffix}'



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Helpful commands for working with a Git repository."""

from datetime import datetime
import logging
from pathlib import Path
import re
import shlex
import subprocess
from typing import Collection, Iterable, Pattern

from pw_cli.plural import plural
from pw_cli.tool_runner import ToolRunner

_LOG = logging.getLogger(__name__)

TRACKING_BRANCH_ALIAS = '@{upstream}'
_TRACKING_BRANCH_ALIASES = TRACKING_BRANCH_ALIAS, '@{u}'
_NON_TRACKING_FALLBACK = 'HEAD~10'


class GitError(Exception):
    """A Git-raised exception."""

    def __init__(
        self, args: Iterable[str], message: str, returncode: int
    ) -> None:
        super().__init__(f'`git {shlex.join(args)}` failed: {message}')
        self.returncode = returncode


class _GitTool:
    def __init__(self, tool_runner: ToolRunner, working_dir: Path) -> None:
        self._run_tool = tool_runner
        self._working_dir = working_dir

    def __call__(self, *args, **kwargs) -> str:
        cmd = ('-C', str(self._working_dir), *args)
        proc = self._run_tool(tool='git', args=cmd, **kwargs)

        if proc.returncode != 0:
            if not proc.stderr:
                err = '(no output)'
            else:
                err = proc.stderr.decode().strip()
            raise GitError((str(s) for s in cmd), err, proc.returncode)

        return '' if not proc.stdout else proc.stdout.decode().strip()


class GitRepo:
    """Represents a checked out Git repository that may be queried for info."""

    def __init__(self, root: Path, tool_runner: ToolRunner):
        self._root = root.resolve()
        self._git = _GitTool(tool_runner, self._root)

    def tracking_branch(
        self,
        fallback: str | None = None,
    ) -> str | None:
        """Returns the tracking branch of the current branch.

        Since most callers of this function can safely handle a return value of
        None, suppress exceptions and return None if there is no tracking
        branch.

        Returns:
          the remote tracking branch name or None if there is none
        """

        # This command should only error out if there's no upstream branch set.
        try:
            return self._git(
                'rev-parse',
                '--abbrev-ref',
                '--symbolic-full-name',
                TRACKING_BRANCH_ALIAS,
            )

        except GitError:
            return fallback

    def current_branch(self) -> str | None:
        """Returns the current branch, or None if it cannot be determined."""
        try:
            return self._git('rev-parse', '--abbrev-ref', 'HEAD')
        except GitError:
            return None

    def _ls_files(self, pathspecs: Collection[Path | str]) -> Iterable[Path]:
        """Returns results of git ls-files as absolute paths."""
        for file in self._git('ls-files', '--', *pathspecs).splitlines():
            full_path = self._root / file
            # Modified submodules will show up as directories and should be
            # ignored.
            if full_path.is_file():
                yield full_path

    def _diff_names(
        self, commit: str, pathspecs: Collection[Path | str]
    ) -> Iterable[Path]:
        """Returns paths of files changed since the specified commit.

        All returned paths are absolute file paths.
        """
        for file in self._git(
            'diff',
            '--name-only',
            '--diff-filter=d',
            commit,
            '--',
            *pathspecs,
        ).splitlines():
            full_path = self._root / file
            # Modified submodules will show up as directories and should be
            # ignored.
            if full_path.is_file():
                yield full_path

    def list_files(
        self,
        commit: str | None = None,
        pathspecs: Collection[Path | str] = (),
    ) -> list[Path]:
        """Lists files modified since the specified commit.

        If ``commit`` is not found in the current repo, all files in the
        repository are listed.

        Arugments:
            commit: The Git hash to start from when listing modified files
            pathspecs: Git pathspecs use when filtering results

        Returns:
            A sorted list of absolute paths.
        """

        if commit in _TRACKING_BRANCH_ALIASES:
            commit = self.tracking_branch(fallback=_NON_TRACKING_FALLBACK)

        if commit:
            try:
                return sorted(self._diff_names(commit, pathspecs))
            except GitError:
                _LOG.warning(
                    'Error comparing with base revision %s of %s, listing all '
                    'files instead of just changed files',
                    commit,
                    self._root,
                )

        return sorted(self._ls_files(pathspecs))

    def has_uncommitted_changes(self) -> bool:
        """Returns True if this Git repo has uncommitted changes in it.

        Note: This does not check for untracked files.

        Returns:
            True if the Git repo has uncommitted changes in it.
        """

        # Refresh the Git index so that the diff-index command will be accurate.
        # The `git update-index` command isn't reliable when run in parallel
        # with other processes that may touch files in the repo directory, so
        # retry a few times before giving up. The hallmark of this failure mode
        # is the lack of an error message on stderr, so if we see something
        # there we can assume it's some other issue and raise.
        retries = 6
        for i in range(retries):
            try:
                self._git(
                    'update-index',
                    '-q',
                    '--refresh',
                    pw_presubmit_ignore_dry_run=True,
                )
            except subprocess.CalledProcessError as err:
                if err.stderr or i == retries - 1:
                    raise
                continue

        try:
            self._git(
                'diff-index',
                '--quiet',
                'HEAD',
                '--',
                pw_presubmit_ignore_dry_run=True,
            )
        except GitError as err:
            # diff-index exits with 1 if there are uncommitted changes.
            if err.returncode == 1:
                return True

            # Unexpected error.
            raise

        return False

    def root(self) -> Path:
        """The root file path of this Git repository.

        Returns:
            The repository root as an absolute path.
        """
        return self._root

    def list_submodules(
        self, excluded_paths: Collection[Pattern | str] = ()
    ) -> list[Path]:
        """Query Git and return a list of submodules in the current project.

        Arguments:
            excluded_paths: Pattern or string that match submodules that should
                not be returned. All matches are done on posix-style paths
                relative to the project root.

        Returns:
            List of "Path"s which were found but not excluded. All paths are
            absolute.
        """
        discovery_report = self._git(
            'submodule',
            'foreach',
            '--quiet',
            '--recursive',
            'echo $toplevel/$sm_path',
        )
        module_dirs = [Path(line) for line in discovery_report.split()]

        for exclude in excluded_paths:
            if isinstance(exclude, Pattern):
                for module_dir in reversed(module_dirs):
                    if exclude.fullmatch(
                        module_dir.relative_to(self._root).as_posix()
                    ):
                        module_dirs.remove(module_dir)
            else:
                for module_dir in reversed(module_dirs):
                    print(f'not regex: {exclude}')
                    if exclude == module_dir.relative_to(self._root).as_posix():
                        module_dirs.remove(module_dir)

        return module_dirs

    def commit_message(self, commit: str = 'HEAD') -> str:
        """Returns the commit message of the specified commit.

        Defaults to ``HEAD`` if no commit specified.

        Returns:
            Commit message contents as a string.
        """
        return self._git('log', '--format=%B', '-n1', commit)

    def commit_author(self, commit: str = 'HEAD') -> str:
        """Returns the author of the specified commit.

        Defaults to ``HEAD`` if no commit specified.

        Returns:
            Commit author as a string.
        """
        return self._git('log', '--format=%ae', '-n1', commit)

    def commit_date(self, commit: str = 'HEAD') -> datetime:
        """Returns the datetime of the specified commit.

        Defaults to ``HEAD`` if no commit specified.

        Returns:
            Commit datetime as a datetime object.
        """
        return datetime.fromisoformat(
            self._git('log', '--format=%aI', '-n1', commit)
        )

    def commit_hash(
        self,
        commit: str = 'HEAD',
        short: bool = True,
    ) -> str:
        """Returns the hash associated with the specified commit.

        Defaults to ``HEAD`` if no commit specified.

        Returns:
            Commit hash as a string.
        """
        args = ['rev-parse']
        if short:
            args += ['--short']
        args += [commit]
        return self._git(*args)

    def commit_change_id(self, commit: str = 'HEAD') -> str | None:
        """Returns the Gerrit Change-Id of the specified commit.

        Defaults to ``HEAD`` if no commit specified.

        Returns:
            Change-Id as a string, or ``None`` if it does not exist.
        """
        message = self.commit_message(commit)
        regex = re.compile(
            'Change-Id: (I[a-fA-F0-9]+)',
            re.MULTILINE,
        )
        match = regex.search(message)
        return match.group(1) if match else None

    def commit_parents(self, commit: str = 'HEAD') -> list[str]:
        args = ['log', '--pretty=%P', '-n', '1', commit]
        return self._git(*args).split()


def find_git_repo(path_in_repo: Path, tool_runner: ToolRunner) -> GitRepo:
    """Tries to find the root of the Git repo that owns ``path_in_repo``.

    Raises:
        GitError: The specified path does not live in a Git repository.

    Returns:
        A GitRepo representing the the enclosing repository that tracks the
        specified file or folder.
    """
    git_tool = _GitTool(
        tool_runner,
        path_in_repo if path_in_repo.is_dir() else path_in_repo.parent,
    )
    root = Path(
        git_tool(
            'rev-parse',
            '--show-toplevel',
        )
    )

    return GitRepo(root, tool_runner)


def is_in_git_repo(p: Path, tool_runner: ToolRunner) -> bool:
    """Returns true if the specified path is tracked by a Git repository.

    Returns:
        True if the specified file or folder is tracked by a Git repository.
    """
    try:
        find_git_repo(p, tool_runner)
    except GitError:
        return False

    return True


def _describe_constraints(
    repo: GitRepo,
    working_dir: Path,
    commit: str | None,
    pathspecs: Collection[Path | str],
    exclude: Collection[Pattern[str]],
) -> Iterable[str]:
    if not repo.root().samefile(working_dir):
        yield (
            'under the '
            f'{working_dir.resolve().relative_to(repo.root().resolve())}'
            ' subdirectory'
        )

    if commit in _TRACKING_BRANCH_ALIASES:
        commit = repo.tracking_branch()
        if commit is None:
            _LOG.warning(
                'Attempted to list files changed since the remote tracking '
                'branch, but the repo is not tracking a branch'
            )

    if commit:
        yield f'that have changed since {commit}'

    if pathspecs:
        paths_str = ', '.join(str(p) for p in pathspecs)
        yield f'that match {plural(pathspecs, "pathspec")} ({paths_str})'

    if exclude:
        yield (
            f'that do not match {plural(exclude, "pattern")} ('
            + ', '.join(p.pattern for p in exclude)
            + ')'
        )


def describe_git_pattern(
    working_dir: Path,
    commit: str | None,
    pathspecs: Collection[Path | str],
    exclude: Collection[Pattern],
    tool_runner: ToolRunner,
    project_root: Path | None = None,
) -> str:
    """Provides a description for a set of files in a Git repo.

    Example:

        files in the pigweed repo
        - that have changed since origin/main..HEAD
        - that do not match 7 patterns (...)

    The unit tests for this function are the source of truth for the expected
    output.

    Returns:
        A multi-line string with descriptive information about the provided
        Git pathspecs.
    """
    repo = find_git_repo(working_dir, tool_runner)
    constraints = list(
        _describe_constraints(repo, working_dir, commit, pathspecs, exclude)
    )

    name = repo.root().name
    if project_root and project_root != repo.root():
        name = str(repo.root().relative_to(project_root))

    if not constraints:
        return f'all files in the {name} repo'

    msg = f'files in the {name} repo'
    if len(constraints) == 1:
        return f'{msg} {constraints[0]}'

    return msg + ''.join(f'\n    - {line}' for line in constraints)
